

--[[

-- Performance?
PlateHandler:SetFrameStrata("TOOLTIP")

-- Consider using UI Scale with the Hub?

Compatibility Mode:

The key differences with Metasequoia:
	- Each nameplate gets its own OnUpdate handler
	- Each nameplates gets its own queue
	- The main queue and onupdate handler is maintained, creating a hybrid system
	- Mouseovers will be handled by the main handler?
	- Simpler Update/Gather functions (UpdateUnitIdentity, UpdateUnitCondition, UpdateUnitContext, Etc.)
	- Optional Anchoring Tech

Nameplate Addon Parts

*1. Plate Detection
*2. Init Function to patch-in our textures and code
*3. Plate.OnUpdate handler to detect mouseover events, alpha, etc.
~5. Data Gathering Functions
6. Styler
7. Event Handlers for Target, Level, Threat, etc. Gamestate changes.
8. Indicator/Delegate Functions  (Indicator functions should be QUEUEABLE)
9. Widget Integration

* Should Widgets have their own carrier?

--
Metasequoia Core
PLAYER_ENTERING_WORLD
PLAYER_TARGET_CHANGED
PLAYER_REGEN_ENABLED
PLAYER_REGEN_DISABLED
RAID_TARGET_UPDATE
UNIT_THREAT_SITUATION_UPDATE
UNIT_LEVEL

PLAYER_CONTROL_LOST
PLAYER_CONTROL_GAINED
UNIT_FACTION

GameState table?

--]]


-- Test Stuff
local BlizzardTheme = {}



-----------------

Metasequoia = {}

local AddonName, Namespace = ...
local MetasequoiaCore = CreateFrame("Frame")
local Nameplates = {}
local VisibleNameplates = {}
local enableTransitions = true
local EMPTY_TEXTURE =  "Interface\\Addons\\TidyPlates\\Media\\Empty"

-----------------------------------------------------------------------------------------------
-- Declarations
-----------------------------------------------------------------------------------------------
-- New Plate Init Function Declarations
local InitFrame, InitVisible

-- Plate Frame Handler Declarations
local OnUpdateFrame, OnShowFrame, OnHideFrame, OnEventFrame
local OnShowPlate, OnHidePlate, OnChangePlate
local EventFrameFunctions = {}

-- Data Gather Declarations
local UpdateUnitIdentity, UpdateUnitCondition, UpdateUnitActivity, UpdateUnitContext

-- Queue Handling Functions
local QueueFunction, QueueFunctionAll

-- Indicators & Style Declarations
local UpdateFrameStyle, UpdateFrameIndicators

-----------------------------------------------------------------------------------------------
-- Plate Detection: Searches for new nameplates, and passes them to InitFrame()
-----------------------------------------------------------------------------------------------

do
	local WorldFrame = WorldFrame
	
	local function IsNameplate(frame)		
		local bars = frame:GetChildren()
		if bars then 
			local border = select(2, bars:GetRegions() )
			return	border and
				border.GetTexture and
				border:GetTexture() == "Interface\\Tooltips\\Nameplate-Border"
		end
	end
	
	local lastUpdate, childCount = 0, 0
	
	local function SearchForNewPlates(self, t)
		lastUpdate = lastUpdate + t

		if lastUpdate >= .03 then		-- kui is .1
			lastUpdate = 0
			local n = select('#', WorldFrame:GetChildren())

			if n ~= childCount then
				local i, f

				for i = 1, n do
					f = select(i, WorldFrame:GetChildren())
					if IsNameplate(f) and not f.Metasequoia then
						
						InitFrame(f)
					end
				end

				childCount = n
			end
		end
	end
	
	MetasequoiaCore:SetFrameStrata("TOOLTIP")
	MetasequoiaCore:SetScript("OnUpdate", SearchForNewPlates); print("DefaultCore Enabled")
end

-----------------------------------------------------------------------------------------------
-- InitFrame: Hook Plate & Add Custom Frames, Passes off to the Show() handler
-----------------------------------------------------------------------------------------------

-- InitFrame: Called by detector function; Creates the complimentary frame; Hooks functions 
function InitFrame(plate)
	Nameplates[plate] = true
	
	-- Get References to the Blizzard Plate Children and Regions
	local overlayChild, nameTextChild  = plate:GetChildren()
	local healthBar, castBar			 = overlayChild:GetChildren()
	
	local threatGlowRegion, overlayRegion,
		highlightRegion, levelTextRegion, 
		bossIconRegion, raidIconRegion,
		eliteIconRegion
		 = overlayChild:GetRegions()

	local healthbarTexture = healthBar:GetRegions()
	
	local castbarTexture, castbarOverlay, shieldedRegion, 
		spellIconRegion, spellTextRegion, 
		spellShadowRegion 
		= castBar:GetRegions()	
	

	local nameTextRegion	= nameTextChild:GetRegions()
	
	local Blizzard = {
		["threatGlowRegion"] = threatGlowRegion, 
		["overlayRegion"] = overlayRegion,
		["highlightRegion"] = highlightRegion,
		["levelTextRegion"] = levelTextRegion, 
		["bossIconRegion"] = bossIconRegion, 
		["raidIconRegion"] = raidIconRegion,
		["eliteIconRegion"] = eliteIconRegion,
		["castbarOverlay"] = castbarOverlay, 
		["shieldedRegion"] = shieldedRegion,
		["spellIconRegion"] = spellIconRegion,
		["nameTextRegion"] = nameTextRegion,
		["healthBar"] = healthBar, 
		["castBar"] = castBar,	
		["overlayChild"] = overlayChild,
		["spellTextRegion"] = spellTextRegion,
		["spellShadowRegion"] = spellShadowRegion,
	}
	
	-- Hide Original Elements
	highlightRegion:SetTexture(nil)
	--bossIconRegion:SetTexture(nil)				-- Reuse ME  
	shieldedRegion:SetTexture(nil)
	castbarOverlay:SetTexture(nil)
	threatGlowRegion:SetTexture(nil)               
	overlayRegion:SetTexCoord( 0, 0, 0, 0 )             
	eliteIconRegion:SetTexCoord( 0, 0, 0, 0 )                             
	--levelTextRegion:SetWidth( .1 )         		-- Reuse ME   ?         
	spellIconRegion:SetTexCoord( 0, 0, 0, 0 )         
	--raidIconRegion:SetAlpha( 0 )   				-- Reuse ME                   
	
	-- Patch 5.3+
	spellTextRegion:Hide()
	spellShadowRegion:Hide()
	-- Patch 5.3+
	
	healthBar:Hide()		-- Interesting... It still updates while hidden?
	nameTextRegion:Hide()
	--castBar:SetStatusBarTexture(EMPTY_TEXTURE)
	castbarTexture:Hide()
	
	-- Create Metasequoia Frame
	local frame = CreateFrame("Frame", nil, WorldFrame)
	frame:SetFrameStrata("BACKGROUND")
	frame.lastUpdate = 0
	frame.Plate = plate
	plate.Metasequoia = true
	
	plate.MetasequoiaFrame = frame
	healthBar.MetasequoiaFrame = frame
	overlayChild.MetasequoiaFrame = frame
	frame.ParentAnchor = overlayChild
	
	-- Tidy Plates Compatibility --
		plate.extended = frame
	-- Tidy Plates Compatibility --
	
	frame.Queue, frame.Unit, frame.UnitCache, frame.Visible = {}, {}, {}, {}
	frame.Blizzard, frame.Style  = 	Blizzard, {}
	
	frame:SetScript("OnShow", OnShowFrame)
	frame:SetScript("OnHide", OnHideFrame)
	frame:SetScript("OnUpdate", OnUpdateFrame)
	frame:SetScript("OnEvent", OnEventFrame)
	
	for event in pairs(EventFrameFunctions) do MetasequoiaCore:RegisterEvent(event) end
	
	plate:HookScript("OnShow", OnShowPlate )
	plate:HookScript("OnHide", OnHidePlate )
	
	--overlayChild:HookScript("OnShow", OnShowPlate )
	--overlayChild:HookScript("OnHide", OnHidePlate )
	--overlayChild:Hide()
	--healthBar:HookScript("OnShow", OnShowPlate )
	--healthBar:HookScript("OnHide", OnHidePlate )
	
	healthBar:HookScript("OnValueChanged", OnChangePlate ) 
	healthBar:HookScript("OnMinMaxChanged", OnChangePlate )
	
	-- Position
	frame:SetPoint("CENTER", overlayRegion, "CENTER")
	--frame:SetPoint("CENTER", WorldFrame, "CENTER")
	frame:SetSize(100, 25)

	-- Alpha
	frame.requestedAlpha, frame.currentAlpha = 0, 0
	frame:SetAlpha(0)
	
	-- Scale
	frame.requestedScale, frame.currentScale = 1, 1
	
	-- Graphics, Text & Textures
	local visible = frame.Visible
	
	healthBar = CreateFrame("StatusBar", nil, frame)
	castBar = CreateFrame("StatusBar", nil, frame)
	
	visible.healthbar = healthBar
	visible.castbar = castBar
	
	-- Health Bar/Lowest Frame
	visible.healthborder = healthBar:CreateTexture(nil, "ARTWORK")
	visible.threatborder = healthBar:CreateTexture(nil, "ARTWORK")
	visible.highlight = healthBar:CreateTexture(nil, "OVERLAY")
	-- Extended/Tidy Plates Root Frame
	--visible.raidicon = healthBar:CreateTexture(nil, "ARTWORK")
	visible.eliteicon = healthBar:CreateTexture(nil, "OVERLAY")
	--visible.skullicon = healthBar:CreateTexture(nil, "OVERLAY")
	visible.target = healthBar:CreateTexture(nil, "BACKGROUND")       -- ARTWORK
	-- Cast Bar Frame
	visible.castborder = castBar:CreateTexture(nil, "ARTWORK")
	visible.castnostop = castBar:CreateTexture(nil, "ARTWORK")
	visible.spellicon = castBar:CreateTexture(nil, "OVERLAY")
	
	--visible.raidicon:SetTexture("Interface\\TargetingFrame\\UI-RaidTargetingIcons")	
	visible.highlight:SetAllPoints(visible.healthborder)
	visible.highlight:SetBlendMode("ADD")  

	-- Reusing Frames
	visible.skullicon, visible.raidicon, visible.level = bossIconRegion, raidIconRegion, levelTextRegion		-- Reuseing
	
	raidIconRegion:SetParent(healthBar)
	levelTextRegion:SetParent(healthBar)
	bossIconRegion:SetParent(frame)
	--local i, v; for i, v in pairs(visible) do v:SetNonBlocking(true) end

	-- Formatted Text
	-------------------------------------------------------------------------------------
	visible.customtext = frame:CreateFontString(nil, "OVERLAY")
	visible.name  = healthBar:CreateFontString(nil, "OVERLAY")
	--visible.level = healthBar:CreateFontString(nil, "OVERLAY")
	-- Cast Bar Frame
	visible.spelltext = castBar:CreateFontString(nil, "OVERLAY")
	
	castBar:Hide()
	
	-- And, finally, Show the Frame
	OnShowFrame(frame)
end

	
-----------------------------------------------------------------------------------------------
-- Plate Frame Handlers
-----------------------------------------------------------------------------------------------

-- OnShowFrame:
function OnShowFrame(frame)
	local b = frame.Blizzard
	local v = frame.Visible
	wipe(frame.UnitCache)
	wipe(frame.Unit)
	
	UpdateUnitContext(frame)
	UpdateUnitIdentity(frame)
	UpdateUnitCondition(frame)
	
	QueueFunction( frame, UpdateFrameIndicators)
end

-- OnHideFrame: 
function OnHideFrame(frame)
	frame.requestedAlpha = 0
	frame.currentAlpha = 0
	frame:SetAlpha(0)
	
	frame.requestedScale = 1
	frame.currentScale = 1
	frame:SetScale(1)
end

-- OnUpdate Handler: Visible nameplates call OnUpdateFrame() on every frame cycle.
do
	local currentTarget, currentMouseover = nil, nil
	local lastTarget, lastMouseover = nil, nil
	
	-- Helper Function Declarations
	local GetTransition
	
	-- OnUpdateFrame: Called frequently by each visible nameplate
	function OnUpdateFrame(frame, e)
		local unit = frame.Unit
		
		--[[
		-- Throttling
		interval = interval + e
		if interval > 1 then 
			interval = 0
		else return
		end
		--]]
		
		-- Mouseover
		if frame.Blizzard.highlightRegion:IsShown() then
			if not unit.isMouseover then
				unit.isMouseover = true
				QueueFunction(frame, UpdateUnitContext)
			end
		elseif unit.isMouseover then
			unit.isMouseover = false
			QueueFunction(frame, UpdateUnitContext)
		end
		
		-- Target
		if frame.Plate:GetAlpha() > .9 and UnitExists("target") then
			if not unit.isTarget then
				unit.isTarget = true
				QueueFunction(frame, UpdateUnitContext)
			end
		elseif unit.isTarget then
			unit.isTarget = false
			QueueFunction(frame, UpdateUnitContext)
		end
		
		-- Alpha Transition
		if  frame.currentAlpha ~= frame.requestedAlpha then
			frame.currentAlpha = GetTransition(frame.currentAlpha, frame.requestedAlpha, e, 4)
			frame:SetAlpha(frame.currentAlpha)
		end

		-- Scale Transition
		if frame.currentScale ~= frame.requestedScale then
			frame.currentScale =  GetTransition(frame.currentScale, frame.requestedScale, e, 4)
			frame:SetScale(frame.currentScale)
		end
		
		-- Function Queue
		local queue, func = frame.Queue, nil
		
		for func in pairs(queue) do
			--print(frame, GetTime())
			func(frame)
			queue[func] = nil
		end
	end	-- End of OnUpdateFrame

	-- GetTransition: 
	function GetTransition(current, requested, e, rate)
			if enableTransitions then 
				local delta = current - requested
				if delta > (e * rate) then
					return current - (e * rate)
				elseif delta < -(e * rate) then
					return current + (e * rate)
				end
			end
			return requested
	end

end

function OnEventFrame(frame, event, ...)
	if event and EventFrameFunctions[event] then EventFrameFunctions[event](frame, ...) end
end


-- OnShowPlate, OnHidePlate:  Redirects events from the Blizzard Nameplate
function OnShowPlate(plate) plate.MetasequoiaFrame:Show(); VisibleNameplates[plate] = true end
function OnHidePlate(plate) plate.MetasequoiaFrame:Hide() VisibleNameplates[plate] = nil end

-- Health Changes
function OnChangePlate(plate) UpdateUnitCondition(plate.MetasequoiaFrame) end

-- Events
do		  
	function EventFrameFunctions:PLAYER_REGEN_ENABLED(frame)  
		QueueFunction(frame, UpdateUnitIdentity)
		QueueFunction(frame, UpdateUnitCondition)
	end
		
	function EventFrameFunctions:PLAYER_REGEN_DISABLED(frame) 
		QueueFunction(frame, UpdateUnitIdentity)
		QueueFunction(frame, UpdateUnitCondition)
	end
	
	function EventFrameFunctions:RAID_TARGET_UPDATE(frame) 		QueueFunction(frame, UpdateUnitCondition); end
	function EventFrameFunctions:UNIT_THREAT_SITUATION_UPDATE(frame) QueueFunction(frame, UpdateUnitCondition); end
	
	function EventFrameFunctions:UNIT_LEVEL(frame) 			QueueFunction(frame, UpdateUnitIdentity); end
	function EventFrameFunctions:PLAYER_CONTROL_LOST(frame)	QueueFunction(frame, UpdateUnitIdentity); end
	
	EventFrameFunctions.PLAYER_CONTROL_GAINED = EventFrameFunctions.PLAYER_CONTROL_LOST
	EventFrameFunctions.UNIT_FACTION = EventFrameFunctions.PLAYER_CONTROL_LOST	
end

-- Function Queueing
do
	-- QueueFunction:
	function QueueFunction(frame, func)
		if func and frame.Queue then frame.Queue[func] = true end
	end

	-- QueueFunctionAll:
	function QueueFunctionAll(func)
		if func then
			for p in pairs(VisibleNameplates) do
				p.MetasequoiaFrame.Queue[func] = true
			end
		end
	end
end

-----------------------------------------------------------------------------------------------
-- Data Gather
-----------------------------------------------------------------------------------------------
do
	-- Helper function declaration
	local ColorToString, GetUnitReaction, GetUnitAggroStatus, GetUnitCombatStatus, UpdateCache
	local ClassReference = {}
	local RaidIconCoordinate = {}
	
	-- UpdateUnitIdentity: Gathers Low-volatility information
	function UpdateUnitIdentity(frame)
		local unit = frame.Unit
		local bliz = frame.Blizzard
		local viz = frame.Visible
			
		-- Name
		unit.name = bliz.nameTextRegion:GetText()
		
		-- Skull
		unit.isBoss = bliz.bossIconRegion:IsShown()
		
		-- Level
		if unit.isBoss then 
			unit.level = "??"
			unit.levelcolorRed, unit.levelcolorGreen, unit.levelcolorBlue = 1, 0, 0
		else 
			unit.level = bliz.levelTextRegion:GetText() 
			unit.levelcolorRed, unit.levelcolorGreen, unit.levelcolorBlue = bliz.levelTextRegion:GetTextColor()
		end

		-- Elite
		unit.isElite = (bliz.eliteIconRegion:IsShown() or 0) == 1
		
		-- Reaction/Type
		unit.red, unit.green, unit.blue = bliz.healthBar:GetStatusBarColor()
		unit.reaction, unit.type = GetUnitReaction(unit.red, unit.green, unit.blue)
		
		-- Class
		unit.class = ClassReference[ColorToString(unit.red, unit.green, unit.blue)] or "UNKNOWN"
		
		-- ** Instead of Calling for the update, how about letting the OnUpdate function handle it?
		-- ** After the function queue is called, it could check for the frame.needsUpdate flag, and then
		-- ** call the appropriate functions.
		
		-- Check for changes
		if UpdateCache(unit, frame.UnitCache) then
			-- Record the FrameNumber when the Update was called.  If called again, defer...
			-- frame.lastUpdate
			-- frame.
			--frame.QueueUpdateIndicators()
			QueueFunction( frame, UpdateFrameIndicators)
		end
	end
	
	
	-- UpdateUnitCondition: Gathers information about the variable condition of a unit
	function UpdateUnitCondition(frame)
		local unit = frame.Unit
		local bliz = frame.Blizzard

		-- Health
		unit.health = bliz.healthBar:GetValue() or 0
		unit.healthmax = select(2, bliz.healthBar:GetMinMaxValues())
		
		-- Threat
		if InCombatLockdown() then
			unit.threatSituation, unit.threatValue = GetUnitAggroStatus(bliz.threatGlowRegion) 
		else unit.threatSituation = "LOW"; unit.threatValue = 0 end
		
		-- In Combat
		unit.isInCombat = GetUnitCombatStatus(bliz.nameTextRegion:GetTextColor())
		
		-- Raid Marked
		unit.isMarked = bliz.raidIconRegion:IsShown() or false
		if unit.isMarked then 
			ux, uy = bliz.raidIconRegion:GetTexCoord()
			unit.raidIconRegion = RaidIconCoordinate[ux][uy]
		else unit.raidIconRegion = nil end
		
		-- Check for changes
		if UpdateCache(unit, frame.UnitCache) then
			QueueFunction( frame, UpdateFrameIndicators)
		end
	end

	
	-- UpdateUnitActivity: 
	function UpdateUnitActivity(frame)
		-- Casting
	end

	-- UpdateUnitContext: 
	function UpdateUnitContext(frame)
		local unit = frame.Unit
		local bliz = frame.Blizzard
		local vis = frame.Visible
		
		-- Mouseover
		if unit.isMouseover then
			unit.guid = UnitGUID("mouseover") 
			vis.highlight:Show()
		else 
			vis.highlight:Hide()
		end
		
		-- Target
		if unit.isTarget then
			unit.guid = UnitGUID("target") 
			vis.target:Show()
		else 
			vis.target:Hide()
		end
		
		QueueFunction( frame, UpdateFrameIndicators)

	end
	
	
	-- Update Cache
	function UpdateCache(unit, cache)
		local changed = false
		for property, value in pairs(unit) do 
			if cache[property] ~= value then
				cache[property] = value
				changed = true
			end
		end
		return changed
	end
	
	-- ColorToString: Converts a color to a string with a C- prefix
	function ColorToString(r,g,b) return "C"..math.floor((100*r) + 0.5)..math.floor((100*g) + 0.5)..math.floor((100*b) + 0.5) end
	
	-- GetUnitReaction: Determines the reaction, and type of unit from the health bar color
	function GetUnitReaction(red, green, blue)
		if red < .01 then 	-- Friendly
			if blue < .01 and green > .99 then return "FRIENDLY", "NPC"
			elseif blue > .99 and green < .01 then return "FRIENDLY", "PLAYER"
			end
		elseif red > .99 then
			if blue < .01 and green > .99 then return "NEUTRAL", "NPC"
			elseif blue < .01 and green < .01 then return "HOSTILE", "NPC"
			end
		elseif red > .5 and red < .6 then
			if green > .5 and green < .6 and blue > .5 and blue < .6 then return "TAPPED", "NPC" end 	-- .533, .533, .99	-- Tapped Mob
		end
		return "HOSTILE", "PLAYER"
	end

	RaidIconCoordinate[0]		= { [0]		= "STAR", 	[0.25]	= "MOON", 	}
	RaidIconCoordinate[0.25]		= { [0]		= "CIRCLE", 	[0.25]	= "SQUARE",	}
	RaidIconCoordinate[0.5]		= { [0]		= "DIAMOND", 	[0.25]	= "CROSS", 	}
	RaidIconCoordinate[0.75]		= { [0]		= "TRIANGLE",	[0.25]	= "SKULL", 	}
		
	-- Populates the class color lookup table
	for classname, color in pairs(RAID_CLASS_COLORS) do 
		local r, g, b = color.r, color.g, color.b
		local colorstring = ColorToString(r, g, b)
		ClassReference[colorstring] = classname 
	end
	ClassReference["C010060"] = "MONK"
	
	-- Determines if a unit is in combat by checking the name text color
	function GetUnitCombatStatus(r, g, b) return (r > .5 and g < .5) end
	
	-- Determines if a unit is attacking, by looking at aggro glow region
	function GetUnitAggroStatus( region)
		-- High = 1, 0, 0	-- Medium High = 1, .6, 0  -- Medium Low = 1, 1, .47
		if not region:IsShown() then return "LOW", 0 end
		local red, green, blue = region:GetVertexColor()
		if red > 0 then
			if green > 0 then
				if blue > 0 then return "MEDIUM", 1 end
				return "MEDIUM", 2
			end
			return "HIGH", 3
		end
	end
	
end



-----------------------------------------------------------------------------------------------
-- Indicator Update
-----------------------------------------------------------------------------------------------


function UpdateFrameIndicators(frame)
	local v = frame.Visible
	local unit = frame.Unit
	
	frame.Style = BlizzardTheme
	
	-- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	-- Style
	
	UpdateFrameStyle(frame)
	
	
	-- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	-- Indicators
	
	v.name:SetText(unit.name)
	v.name:SetTextColor(unit.red, unit.green, unit.blue)
	
	v.healthbar:SetStatusBarColor(unit.red, unit.green, unit.blue)
	v.healthbar:SetValue(unit.health)
	v.healthbar:SetMinMaxValues(0, unit.healthmax)
	 
	
	-- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	-- DELEGATE TEST
	--v.name:SetText(unit.name.."@"..unit.health.."~"..unit.threatValue)
	--print("UpdateFrameIndicators", unit.name, GetTime())
	--[[
	v.name:SetFont(NAMEPLATE_FONT, 12, "NONE")
	v.name:SetPoint("CENTER", frame, "CENTER")
	v.name:SetSize(200, 32)
	--]]

	
	if unit.isMouseover then
		frame:SetFrameLevel(1)
	elseif unit.isTarget then
		frame:SetFrameLevel(2)
	else
		frame:SetFrameLevel(0)
	end
	
	if (not UnitExists("target")) or unit.isMouseover or unit.isTarget then
		frame.requestedAlpha = 1
	else
		frame.requestedAlpha = .5
	end

	if unit.isTarget then
		frame.requestedScale = 1.5
	else
		frame.requestedScale = 1
	end
end

-- [[
---------------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------
-- II. Frame/Layer Style Appearance Functions:  These functions set the appearance of 
-- specific object types within a style.
---------------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------

local function SetObjectShape(object, width, height) object:SetWidth(width); object:SetHeight(height) end
local function SetObjectFont(object,  font, size, flags) if not object:SetFont(font, size, flags) then object:SetFont("FONTS\\ARIALN.TTF", size or 12, flags) end end
local function SetObjectJustify(object, horz, vert) object:SetJustifyH(horz); object:SetJustifyV(vert) end
local function SetObjectAnchor(object, anchor, anchorTo, x, y) object:ClearAllPoints();object:SetPoint(anchor, anchorTo, anchor, x, y) end
local function SetObjectTexture(object, texture) object:SetTexture(texture) end 
local function SetObjectBartexture(obj, tex, ori, crop) obj:SetStatusBarTexture(tex); obj:SetOrientation(ori); end
local function SetObjectShadow(object, shadow) if shadow then
    object:SetShadowColor(0,0,0, tonumber(shadow) or 1); object:SetShadowOffset(.5, -.5)
                else object:SetShadowColor(0,0,0,0) end
end
-- SetFontGroupObject
local function SetFontGroupObject(object, objectstyle)
	if objectstyle then
		SetObjectFont(object, objectstyle.typeface or "FONTS\\ARIALN.TTF",  objectstyle.size or 12, objectstyle.flags or "NONE")
		SetObjectJustify(object, objectstyle.align or "CENTER", objectstyle.vertical or "BOTTOM")
		SetObjectShadow(object, objectstyle.shadow or 1)
	end
end
-- SetAnchorGroupObject
local function SetAnchorGroupObject(object, objectstyle, anchorTo)
	if objectstyle and anchorTo then
		SetObjectShape(object, objectstyle.width or 128, objectstyle.height or 16) --end
		SetObjectAnchor(object, objectstyle.anchor or "CENTER", anchorTo, objectstyle.x or 0, objectstyle.y or 0)
	end
end
-- SetTextureGroupObject
local function SetTextureGroupObject(object, objectstyle)
	if objectstyle then
		if objectstyle.texture then SetObjectTexture(object, objectstyle.texture or EMPTY_TEXTURE) end
		object:SetTexCoord(objectstyle.left or 0, objectstyle.right or 1, objectstyle.top or 0, objectstyle.bottom or 1)
	end
end
-- SetBarGroupObject
local function SetBarGroupObject(object, objectstyle, anchorTo)
	if objectstyle then
		SetAnchorGroupObject(object, objectstyle, anchorTo)
		SetObjectBartexture(object, objectstyle.texture or EMPTY_TEXTURE, objectstyle.orientation or "HORIZONTAL")
		if objectstyle.backdrop then
			--object:SetBackdropTexture(objectstyle.backdrop)
		end
		--object:SetTexCoord(objectstyle.left, objectstyle.right, objectstyle.top, objectstyle.bottom)
	end
end

---------------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------
-- III. Nameplate Styler: These functions parses the definition table for a nameplate's requested style.
---------------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------
local UpdateStyle
do
	-- UpdateStyle Variables
	local index, content
	local objectstyle, objectname, objectregion, objectenable
	-- Style Property Groups
	local fontgroup = {"name", "level", "spelltext", "customtext"}
	local anchorgroup = {"healthborder", "threatborder", "castborder", "castnostop",
						"name",  "spelltext", "customtext", "level",
						"spellicon", "raidicon", "skullicon", "eliteicon", "target"}		
	local bargroup = {"castbar", "healthbar"}
	local texturegroup = { "castborder", "castnostop", "healthborder", "threatborder", "eliteicon", 
						"skullicon", "highlight", "target", "spellicon", }
	-- UpdateStyle: 
	function UpdateFrameStyle(frame)
		local style = frame.Style
		local visual = frame.Visible
		-- Frame
		SetAnchorGroupObject(frame, style.frame, frame.ParentAnchor)
		
		-- Anchorgroup
		for index = 1, #anchorgroup do 
			objectname = anchorgroup[index]; SetAnchorGroupObject(visual[objectname], style[objectname], frame) 
			objectenable = style[objectname] and style[objectname].show
			--if objectenable then visual[objectname]:Show() else visual[objectname]:Hide() end
		end
		-- Bars
		for index = 1, #bargroup do objectname = bargroup[index]; SetBarGroupObject(visual[objectname], style[objectname], frame) end
		-- Texture
		for index = 1, #texturegroup do objectname = texturegroup[index]; SetTextureGroupObject(visual[objectname], style[objectname]) end
		-- Font Group
		for index = 1, #fontgroup do objectname = fontgroup[index];SetFontGroupObject(visual[objectname], style[objectname]) end
		
		
		
		--[[
		-- Hide Stuff
		if unit.isElite then visual.eliteicon:Hide() else visual.eliteicon:Hide() end
		if unit.isBoss then visual.level:Hide() else visual.skullicon:Hide() end
		if not unit.isTarget then visual.target:Hide() end
		if not unit.isMarked then visual.raidicon:Hide() end
		--]]
	end
	
	--[[
	
function UpdateFrameStyle(frame)
	local v = frame.Visible
	
	-- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	-- STYLER TEST
	v.name:SetFont(NAMEPLATE_FONT, 12, "NONE")
	v.name:SetPoint("CENTER", frame, "CENTER")
	v.name:SetSize(200, 32)
end
	--]]
end
--]]


--[[
-- External
function TidyPlates:ForceUpdate()  end
function TidyPlates:Update() end
function TidyPlates:RequestWidgetUpdate(plate) end
function TidyPlates:RequestDelegateUpdate(plate) end
function TidyPlates:ActivateTheme(theme)
	-- if theme and type(theme) == 'table' then Metasequoia.ActiveThemeTable, activetheme = theme, theme; SetMassQueue(Event_ResetNameplate) end
									
end
function TidyPlates:EnableFadeIn() end
function TidyPlates:DisableFadeIn() end
TidyPlates.NameplatesByGUID, TidyPlates.NameplatesAll, TidyPlates.NameplatesByVisible = GUID, Plates, PlatesVisible

--]]



------------------------------------------------
--	Blizzard-Style (Used for the default/fallback theme)
------------------------------------------------


BlizzardTheme.highlight = {}
BlizzardTheme.highlight.texture = "Interface\\Tooltips\\Nameplate-Glow"
BlizzardTheme.highlight.width = 128
BlizzardTheme.highlight.height = 32

BlizzardTheme.healthborder = {}
BlizzardTheme.healthborder.texture		 =				"Interface\\Tooltips\\Nameplate-Border"
BlizzardTheme.healthborder.width = 128
BlizzardTheme.healthborder.height = 32
BlizzardTheme.healthborder.x = 0
BlizzardTheme.healthborder.y = 0

BlizzardTheme.eliteicon = {}
BlizzardTheme.eliteicon.texture = "Interface\\Tooltips\\EliteNameplateIcon"
BlizzardTheme.eliteicon.width = 32
BlizzardTheme.eliteicon.height = 32
BlizzardTheme.eliteicon.x = 22
BlizzardTheme.eliteicon.y = -8.5
BlizzardTheme.eliteicon.anchor = "RIGHT"
BlizzardTheme.eliteicon.show = true
BlizzardTheme.eliteicon.left = 0
BlizzardTheme.eliteicon.right = .578125
BlizzardTheme.eliteicon.top = 0
BlizzardTheme.eliteicon.bottom = .84375

BlizzardTheme.threatborder = {}
BlizzardTheme.threatborder.texture =			"Interface\\TargetingFrame\\UI-TargetingFrame-Flash"
BlizzardTheme.threatborder.width = 140
BlizzardTheme.threatborder.height = 32
BlizzardTheme.threatborder.x = 0
BlizzardTheme.threatborder.y = -10
BlizzardTheme.threatborder.left = 0
BlizzardTheme.threatborder.right = .555
BlizzardTheme.threatborder.top = .53
BlizzardTheme.threatborder.bottom = .6

BlizzardTheme.castborder = {}
BlizzardTheme.castborder.texture =					"Interface\\Tooltips\\Nameplate-CastBar"
BlizzardTheme.castborder.width = 128
BlizzardTheme.castborder.height = 32
BlizzardTheme.castborder.x = -17
BlizzardTheme.castborder.y = -24

BlizzardTheme.castnostop = {}
BlizzardTheme.castnostop.texture = 				"Interface\\Tooltips\\Nameplate-CastBar-Shield"
BlizzardTheme.castnostop.width = 128
BlizzardTheme.castnostop.height = 32
BlizzardTheme.castnostop.x = -17
BlizzardTheme.castnostop.y = -24

BlizzardTheme.name = {}
BlizzardTheme.name.typeface = NAMEPLATE_FONT
BlizzardTheme.name.size = 12
BlizzardTheme.name.width = 200
BlizzardTheme.name.height = 16
BlizzardTheme.name.x = 0
BlizzardTheme.name.y = 8
BlizzardTheme.name.align = "CENTER"
BlizzardTheme.name.anchor = "CENTER"
BlizzardTheme.name.shadow = true
BlizzardTheme.name.vertical = "BOTTOM"
BlizzardTheme.name.flags = "OUTLINE"

BlizzardTheme.level = {}
BlizzardTheme.level.typeface = NAMEPLATE_FONT
BlizzardTheme.level.size = 9
BlizzardTheme.level.width = 25
BlizzardTheme.level.height = 10
BlizzardTheme.level.x = 15
BlizzardTheme.level.y = -6.5
BlizzardTheme.level.align = "CENTER"
BlizzardTheme.level.anchor = "RIGHT"
BlizzardTheme.level.vertical = "BOTTOM"
BlizzardTheme.level.shadow = true
BlizzardTheme.level.flags = "NONE"

BlizzardTheme.healthbar = {}
BlizzardTheme.healthbar.texture =					 "Interface\\TARGETINGFRAME\\UI-StatusBar"
BlizzardTheme.healthbar.backdrop = 				EMPTY_TEXTURE
BlizzardTheme.healthbar.height = 9
BlizzardTheme.healthbar.width = 102
BlizzardTheme.healthbar.x = -8.5
BlizzardTheme.healthbar.y = -7

BlizzardTheme.castbar = {}
BlizzardTheme.castbar.texture =					"Interface\\TARGETINGFRAME\\UI-StatusBar"
BlizzardTheme.castbar.backdrop = 				EMPTY_TEXTURE
BlizzardTheme.castbar.height = 9
BlizzardTheme.castbar.width = 102
BlizzardTheme.castbar.x = -8.5
BlizzardTheme.castbar.y = -24

BlizzardTheme.spellicon = {}
BlizzardTheme.spellicon.width = 14
BlizzardTheme.spellicon.height = 14
BlizzardTheme.spellicon.x = -25.5
BlizzardTheme.spellicon.y = -24
BlizzardTheme.spellicon.anchor = "LEFT"
BlizzardTheme.spellicon.show = true

BlizzardTheme.raidicon = {}
BlizzardTheme.raidicon.width = 20
BlizzardTheme.raidicon.height = 20
BlizzardTheme.raidicon.x = 0
BlizzardTheme.raidicon.y = 8
BlizzardTheme.raidicon.anchor = "TOP"
BlizzardTheme.raidicon.show = true

BlizzardTheme.skullicon = {}
BlizzardTheme.skullicon.texture = "Interface\\TargetingFrame\\UI-TargetingFrame-Skull"
BlizzardTheme.skullicon.width = 15
BlizzardTheme.skullicon.height = 15
BlizzardTheme.skullicon.x = 53
BlizzardTheme.skullicon.y = -7
BlizzardTheme.skullicon.anchor = "CENTER"
BlizzardTheme.skullicon.show = true

BlizzardTheme.frame = {}
BlizzardTheme.frame.width = 101
BlizzardTheme.frame.height = 45
BlizzardTheme.frame.x = 0
BlizzardTheme.frame.y = 0
BlizzardTheme.frame.anchor = "CENTER"
